from __future__ import unicode_literals
try:
    import markdown
    has_markdown = True
except ImportError:
    print('please install markdown')
    print('pip instlal markdown')
    has_markdown = False
from .doc import SchemaGenerator


def doc_creator(title='',
                url=None,
                patterns=None,
                urlconf=None,
                querystring=''):
    from django.test import RequestFactory
    from rest_framework import compat
    request = RequestFactory()
    request = request.request(QUERY_STRING=compat.urlparse.urlencode(querystring))
    schema = SchemaGenerator(
        title=title, url=url, patterns=patterns, urlconfi=urlconf).get_schema(
            request=request)
    MarkdownDocFactory(schema.content, request=request, title=title)


class MarkdownDocFactory(object):
    def __init__(self, links, request, title):
        self.field_string = '|{field}|{dec}|{param_type}|{paramter_type}|{required}|\n'
        self.request = request
        self.title = title
        self.url_md_template = '''
## {url}
### 方法:{action}
### 介绍:
    {description}
|字段|说明|数据类型|HTTP参数类型|是否必填|
|:--:|:--:|:------:|:------:|:------:|
'''
        self.links = links

    def _create_url_md(self):
        url_md_list = []
        filter = self.request.GET.get('filter')
        if filter:
            self.links = [x for x in self.links if filter in x.url]
        no = self.request.GET.get('no')
        if no:
            self.links = [x for x in self.links if no not in x.url]
        for link in self.links:
            if link.fields:
                try:
                    description = link.description.decode('utf-8')
                except Exception:
                    print(description)
                    description = link.description

                url_title = self.url_md_template.format(
                    url=link.url, action=link.action, description=description)
                for field in link.fields:
                    dec = markdown.markdown(
                        field.description,
                        extensions=['markdown.extensions.tables'])
                    dec_no_enter = "".join(dec.split())
                    field_str = self.field_string.format(
                        field=field.name,
                        dec=dec_no_enter,
                        param_type=field.type,
                        paramter_type=field.location,
                        required='是' if field.required else '否')
                    url_title += field_str
                url_md_list.append(url_title)
        return url_md_list

    def create(self, title):
        if not has_markdown:
            return
        url_md_list = self._create_url_md()
        swager_md = self.title
        md = ''.join(url_md_list)
        swager_md += md
        with open('APIdoc.md', 'wb') as f:
            f.write(swager_md.encode('utf-8'))
        return swager_md



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function
import re

from openapi_codec import encode
from openapi_codec.encode import generate_swagger_object as _generate_swagger_object
from rest_framework.schemas import AutoSchema as Schema
from rest_framework.compat import coreapi
from openapi_codec.encode import _get_links
from django.db import models

from .fields import _get_properties, _callback

RE_PATH = re.compile('\{(\w+)\}')  # extract  /{arg1}/{arg2}  to [arg1, arg2]


class SwaggerSchema(Schema):
    def get_swagger_fields(self, path, method):
        view = getattr(self.view, method.lower(), None)  #APIView
        if not view:
            view = getattr(self.view, self.view.action, None)  # ViewSet
        swagger = getattr(view, '__swagger__',
                          {})  # para_ok_or_400中生成的__swagger__
        path_args = {x for x in RE_PATH.findall(path)}
        parameters = swagger.get('parameters', [])
        fields = tuple(
            coreapi.Field(
                name=x['name'],
                location=(x['name'] in path_args and 'path') or x['in'],
                required=x['name'] in path_args or x.get('required', False),
                description=x['description'],
                type=x['type']) for x in parameters)
        return fields

    def get_link(self, path, method, base_url):
        """
        Return a `coreapi.Link` instance for the given endpoint.
        """
        base_url = None  # 当设置url时，生成的path中应该不带url
        link = super(SwaggerSchema, self).get_link(path, method, base_url)
        swagger_fields = self.get_swagger_fields(path, method)
        nameset = {x.name for x in swagger_fields}
        fields = tuple(
            (x for x in link.fields
             if (x.name not in nameset and x.name != 'id'))) + swagger_fields
        link = coreapi.Link(
            url=link.url,
            action=link.action,
            encoding=link.encoding,
            fields=fields,
            description=link.description)
        link.__serializer__ = getattr(
            self.view, 'serializer_class',
            None)  # Link是不可变的数据结构。。。所以覆盖带下划线的属性,以绕过这个限制
        return link


def wrap_generator(func):
    def inner(document):
        swagger = _generate_swagger_object(document)
        links = _get_links(document)
        dataset = dict(
            serializergeter(link.__serializer__) for _, link, _ in links
            if getattr(link, '__serializer__', None))
        swagger['definitions'] = dataset
        return swagger

    return inner


def _get_serializer_name(serializer):
    if 'Serializer' in serializer.__name__:
        obj_name = serializer.__name__.split('Serializer')[0]
    else:
        obj_name = serializer.__name__
    return obj_name


def serializergeter(serializer):
    obj_name = _get_serializer_name(serializer)

    if issubclass(serializer, models.Model):
        model = serializer
    else:
        if hasattr(serializer.Meta, 'model'):
            model = serializer.Meta.model
        else:  # set user model
            data = serializer('any').data
            properties = _get_properties(data, [])
            return obj_name, properties
    fields = model._meta.fields
    names = (x.name for x in fields)
    properties = {x: _callback(y) for x, y in zip(names, fields)}
    obj = dict(properties=properties, type='object')
    return obj_name, obj


def _get200(link):
    serializer = getattr(link, '__serializer__', None)
    tpl = dict(description='Success')
    if not serializer:
        return tpl
    obj_name = ''
    if hasattr(serializer, '_meta'):
        obj_name = str(serializer._meta.object_name).lower()
    elif hasattr(serializer, 'Meta'):
        obj_name = _get_serializer_name(serializer)
    ref = {"$ref": "#/definitions/{}".format(obj_name)}
    if link.action == 'get' and '{' not in link.url:  # list view
        tpl['schema'] = dict(items=ref, type='array')
    else:
        tpl['schema'] = ref
    return tpl


def _get400(link):
    tpl = dict(description='Bad Request')
    return tpl


def _get403(link):
    tpl = dict(description='Forbidden')
    return tpl


def _get_responses(link):
    """
    Returns minimally acceptable responses object based
    on action / method type.
    """
    response_template = {
        '200': _get200(link),
        '400': _get400(link),
        '403': _get403(link)
    }
    if link.action.lower() == 'delete':
        response_template.pop('200')
        response_template.update({'204': {'description': 'Success'}})

    return response_template


def patch_all():
    import rest_framework
    rest_framework.schemas.AutoSchema = SwaggerSchema
    encode._get_responses = _get_responses
    encode.generate_swagger_object = wrap_generator(_generate_swagger_object)



# encoding: utf-8
from __future__ import unicode_literals


class Result(object):
    def __init__(self, dataset=None, serializer=None):
        self.errors = []
        self._error = self.errors.append
        self.serializer = serializer
        self.dataset = None
        self.msg = None

    def data(self, dataset):
        self.dataset = dataset
        return self

    def error(self, key, value, **kwargs):
        self._error(dict(kwargs, name=key, value=value))
        return self

    def perm(self, reason):
        self.msg = reason
        return self

    def __nonzero__(self):
        return not self.errors

    def response(self):
        raise NotImplementedError

    def __call__(self, status=200, serialize=False, **kwargs):
        return self.response(status=status, serialize=serialize, **kwargs)


class Valid(object):
    def __init__(self, method, **kwargs):
        self.method = method
        self.status = 200
        self.msg = None
        self.kwargs = kwargs

    def __str__(self):
        return '<Valid: %s>' % self.method

    def __repr__(self):
        return '<Valid: %s>' % self.method

    def __call__(self, *args, **kwargs):
        return getattr(self, self.method)(*args, **kwargs)


class MethodProxy(object):
    kwargs = {}

    def __init__(self, valid_class=None):
        self.valid_class = valid_class

    def __call__(self, *args, **kwargs):
        self.kwargs = kwargs
        return self

    def __getattr__(self, key):
        return self.valid_class(key, **self.kwargs)



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals
import six
from functools import wraps
from uuid import uuid1
from importlib import import_module
from .datastrctures import Valid


def _doc_generater(itemset, func):
    func_name = func.__name__
    locationmap = {
        'get': 'query',
        'list': 'query',
        'retrieve': 'query',
        'update': 'formData',
        'create': 'formData',
        'post': 'formData'
    }
    location = locationmap.get(func_name, 'querystring')

    def defaulter(item):
        """设置默认参数"""

        def type_by_name(name):
            if name.endswith('id'):
                return 'integer'
            if any(x in name for x in ['date', 'start', 'end']):
                return 'date'
            return 'string'

        item.setdefault('in', location)
        required = item['in'] == 'path'
        item.setdefault('method', lambda x: x)
        item['name'] == 'pk' and item.update(
            name='id', type='integer', required=item.get('required', True))
        item.setdefault('required', required)
        item.setdefault('type', type_by_name(item['name']))
        description = item.get('description') or item.get(
            'msg') or 'description'
        msg = getattr(item['method'], 'msg',
                      item.get('msg') or item.get('description')
                      or 'msg')  # method maybe V
        if type(description) in {list, tuple}:
            description = list2mk(
                description, title=['value',
                                    'description'])  # 支持list转化为markdown表格的形式
        elif isinstance(description, dict):
            description = parse_description(
                description, title={'value':
                                    'description'})  # 支持dict转化为markdown表格的形式

        item['description'] = description
        item.setdefault('msg', msg)
        item.setdefault('replace', None)
        method = item['method']
        if not isinstance(method, Valid):  # replace lambda as Valid
            name = str(uuid1()).split('-')[0]
            v = Valid(name)
            setattr(v, name, method)
            item['method'] = v
        return item

    try:
        doc = func.__doc__.decode('utf8')
    except:
        doc = func.__doc__
    title = doc and doc.strip('\n').strip() or func.__name__

    parameters = [defaulter(x) for x in itemset]
    swagger = dict(title=title, parameters=parameters)
    return swagger


def default_data_method(request):
    return request.GET or request.data


def para_ok_or_400(itemset):
    """
    验证参数值, 参数不对则返回400, 若参数正确则返回验证后的值, 并且根据itemset中的值，来生成func的__doc__
    name: 需要校验的参数名称
    method: 校验方法, 校验成功时， 则返回校验方法后的校验值
    required: 是否可以为空
    msg: 校验失败返回的错误消息
    replace: 校验正确后返回的值对应的key
    description: 对参数的描述
    in: path, querystring, formData
    type: 参数类型(string, integer), 可以根据参数名称来确定, user_id(int), start_date(date), string
    """

    def decorator(func):
        from django.conf import settings
        swagger = _doc_generater(itemset, func)
        data_method = default_data_method
        if getattr(settings, 'PARAER_DATA_METHOD', ''):
            data_method = import_module(settings.PARAER_DATA_METHOD,
                                        default_data_method)  # 获取data的方法

        def wrapper(cls, request, *args, **kwargs):
            paramap = dict(kwargs)
            paramap.setdefault(
                'id', kwargs.get('pk', None)
            )  # Serializer fields中生成的为id 这个key， 但是django解析url中为 pk这个pk，为了不在文档中生成id 和pk这两个field， 所以都统一用id这个key， 那么在itemset中也写id这个key
            data = data_method(request)
            paramap.update({x: y for x, y in data.items()})
            result = cls.result_class()  # 继承与Result类
            for item in itemset:
                name, v, required, msg, replace = [
                    item[x]
                    for x in ['name', 'method', 'required', 'msg', 'replace']
                ]
                value = None  # 与 '' 区别
                para = paramap.get(name)
                if required and para in (None,
                                         ''):  # 如果是post方法并且传参是json的话，para可能为0
                    result.error(name, 'required')
                if para is not None:
                    if para:
                        try:
                            value = v(para)
                        except Exception:
                            if settings.DEBUG:
                                from traceback import print_exc
                                print_exc()
                        msg = v.msg or msg
                        if v.status == 403:  # 权限错误时直接返回错误
                            return result.perm(name, msg)(status=v.status)
                        if value in (None, False):
                            result.error(name, msg)
                    kwargs.update({
                        replace or name: value or para
                    })  # method 返回了非布尔值则更新kwargs
            if not result:
                return result(status=400)
            return func(cls, request, *args, **kwargs)

        wrapper.__swagger__ = swagger
        wrapper.__name__ = func.__name__
        wrapper.__doc__ = func.__doc__
        return wrapper

    return decorator


def perm_ok_or_403(itemset):
    """验证参数值, 参数不对则返回400, 若参数正确则返回验证后的值"""

    def decorator(func):
        @wraps(func)
        def wrapper(cls, request, *args, **kwargs):
            for item in itemset:
                before, method, reason = item.get('before'), item[
                    'method'], item['reason']
                before and before(request, kwargs)
                perm = method(request, kwargs)
                if not perm:
                    return cls.result_class().perm(reason=reason)(status=403)
            return func(cls, request, *args, **kwargs)

        return wrapper

    return decorator


def _make(index, data, length):
    if index == 0:
        return ''.join([
            join(data, length), '\n',
            join(((x * '-') for x in length), length)
        ])
    return join(data, length)


def join(data, length):  # 填充空格
    try:
        row = (''.join((x, (length[index] - len(x)) * ' '))
               for index, x in enumerate(data))
        row = '|'.join(row)
    except UnicodeDecodeError:
        print('UnicodeDecodeError!!!')
        print(data)
        return ''
    return ''.join(('|', row, '|'))


def list2mk(dataset, title=None):
    if not dataset:
        return ''
    txt = ''  # 支持这种形式 ["description", [()]]
    if isinstance(dataset[0], six.string_types):
        txt = dataset[0]
        dataset = dataset[1:]
        dataset = [(str(x[0]), str(x[1])) for x in dataset]  # 为了把数字转为字符串
    dataset = list(dataset)  # just shallow copy
    title and dataset.insert(0, title)
    length = [len(x) for x in dataset[0]]  # assert title is [str, str]
    mk = '\n'.join(
        _make(index, data, length) for index, data in enumerate(dataset))
    if txt:
        mk = '\n'.join((txt, '', mk))
    return mk


def parse_description(dataset, title):
    dataset = {
        isinstance(x, int) and str(x) or x: y
        for x, y in dataset.items()
    }  # 吧数字转为字符串
    if isinstance(dataset, list):
        return ''
    txt = dataset.pop('description', '')
    length = [max(map(len, x)) for x in [dataset, dataset.values()]]  # 取出最大长度

    blank = join(((x * '-') for x in length), length)
    title = list(title.items())[0]
    title = '\n'.join((join(title, length), blank))
    mk = '\n'.join(join(data, length) for data in dataset.items())
    mk = '\n'.join((title, mk))
    if txt:
        mk = '\n'.join((txt, '\n', mk))
    return mk
